#!/usr/bin/env python
"""
Name: Rio Atmadja
Date: 11 May 2020
Description: Contasts variables for this project
"""
from typing import List, Dict
DEFAULT_CLF: List[str] = ['logreg', 'knn', 'dtree']
DEFAULT_RGR: List[str] = ['linreg']
MONTHS: Dict = {
    "January":"01",
    "February":"02",
    "March":"03",
    "April":"04",
    "May":"05",
    "June":"06",
    "July":"07",
    "August":"08",
    "September":"09",
    "October":"10",
    "November":"11",
    "December":"12"
}